{************************************************}
{                                                }
{   Extended visual components                   }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.Inc}
unit MultiSel;
interface

uses Drivers, Views, Objects, Dialogs, Register;

type
  PMultiSelBox = ^TMultiSelBox;
  TMultiSelBox = object(TListBox)
    SelectChar:Char;
    constructor Init(var Bounds: TRect; ANumCols: Word; AScrollBar:
      PScrollBar; AChar:Char);
    constructor Load(var S: TStream);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure SelectAll;
    procedure DeselectAll;
    procedure InvertSelection;
    procedure SetSelected(Item:integer; Enable:boolean);virtual;
    function GetSelected(Item:Integer):boolean;virtual;
    procedure DrawItem(Item, ColWidth:Integer; var B);virtual;
    procedure Store(var S: TStream);
    procedure Lock;
    procedure UnLock;
    procedure Update;
  private
    LockCount:integer;
  end;

const
  RMultiSelBox: TStreamRec = (
    ObjType: rcMultiSelBox;
    VmtLink: Ofs(TypeOf(TMultiSelBox)^);
    Load: @TMultiSelBox.Load;
    Store: @TMultiSelBox.Store
  );

procedure RegistersMultiSelect;
implementation

constructor TMultiSelBox.Init(var Bounds: TRect; ANumCols: Word; AScrollBar:
      PScrollBar; AChar:Char);
begin
  inherited Init( Bounds, ANumCols, AScrollBar);
  SelectChar:=AChar;
  ComponentType:=ctMultiSelBox;
end;

procedure TMultiSelBox.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(SelectChar,SizeOf(Char));
end;

constructor TMultiSelBox.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(SelectChar,SizeOf(Char));
end;


procedure TMultiSelBox.HandleEvent(var Event: TEvent);
begin
  if (((Event.What = evMouseDown) and (Event.Buttons=mbRightButton)) or
     ((Event.What = evKeyDown) and (Event.KeyCode = kbIns))) then
  begin
    SetSelected(Focused,not GetSelected(Focused));
    if (Event.What = evKeyDown) and (Event.KeyCode = kbIns) then
       FocusItemNum(Focused+1);
    DrawView;
  end;
  if (Event.What = evKeyDown) then
  case Event.CharCode of
      '+': SelectAll;
      '-': DeselectAll;
      '*': InvertSelection;
  end;
  inherited HandleEvent(Event);
end;

procedure TMultiSelBox.SelectAll;
var i: Integer;
begin
  if List <> nil then
  begin
    Lock;
    for i := 0 to List^.Count - 1 do SetSelected(i,true);
    UnLock;
  end;
end;

procedure TMultiSelBox.DeselectAll;
var i: Integer;
begin
  if List <> nil then
  begin
    Lock;
    for i := 0 to List^.Count - 1 do SetSelected(i,false);
    UnLock;
  end
end;

procedure TMultiSelBox.InvertSelection;
var i: Integer;
begin
  if List <> nil then
  begin
    Lock;
    for i := 0 to List^.Count - 1 do SetSelected(i,not GetSelected(i));
    UnLock;
  end
end;

procedure TMultiSelBox.SetSelected(Item:integer; Enable:boolean);
var
  PS: PString;
begin
  if List <> nil then
  begin
    PS := List^.At(Item);
    if Enable then PS^[1] := SelectChar
    else PS^[1] := ' ';
    Update;
  end;
end;

function TMultiSelBox.GetSelected(Item:Integer):boolean;
begin
  if List <> nil then  GetSelected:=PString(List^.At(item))^[1]=SelectChar
  else GetSelected:=false;
end;

procedure TMultiSelBox.Lock;
begin
  inc(LockCount);
end;

procedure TMultiSelBox.UnLock;
begin
  if LockCount>0 then Dec(LockCount);
  if LockCount=0 then DrawView;
end;

procedure TMultiSelBox.Update;
begin
  if LockCount=0 then DrawView;
end;

procedure TMultiSelBox.DrawItem(Item, ColWidth:Integer; var B);
var i:integer;
    C:byte;
    B1: TDrawBuffer absolute B;
begin
  if GetSelected(Item) and (Item<>Focused) then
  begin
    C:=GetColor(6);
    for i:=0 to ColWidth-2 do WordRec(B1[i]).Hi:=C;
  end;
end;

procedure RegistersMultiSelect;
begin
  RegisterType( RMultiSelBox);
end;

end.