unit
  SPanel; { Simply navigation panel }
  { by Gosha Sakovich aka 2:5020/2546@fidonet.org }

interface

uses
  views, objects, drivers, app
  {$IFDEF VirtualPascal}, use32{$ENDIF}
  ;

type
  PSPanelItem = ^TSPanelItem;
  TSPanelItem = record
    Next: PSPanelItem;
    Text: PString;
    Command: Word;
    {Disabled: Boolean;}
  end;


  { Palette layout }
  { 1 = Normal text        }
  { 2 = Disabled text      }
  { 3 = Normal Selection   }
  { 4 = Disabled Selection }
  { 5 = Pressed            }
  { 6 = Disabled Pressed   }

  PSPanel = ^TSPanel;
  TSPanel = object(TView)
    Items: PSPanelItem;
    constructor Init(var Bounds: TRect; SItems: PSPanelItem);
    procedure HandleEvent(var Event: TEvent); virtual;
    procedure Draw; virtual;
    procedure DrawSelect(Selected: PSPanelItem; Pressed: Boolean);
    destructor Done; virtual;
  end;


function NewSPanelItem(AText: String; ACommand: Word; ANext: PSPanelItem): PSPanelItem;


implementation

{ TSPanel }

constructor TSPanel.Init(var Bounds: TRect; SItems: PSPanelItem);
var
  T: PSPanelItem;
  L: Integer;
begin
  inherited Init(Bounds);
  Items := SItems;
  EventMask := EventMask or evMouseMove or evBroadCast;
  Options := Options or ofPreProcess;

  L := 2; T := Items;
  while T <> nil do
    begin
      if T^.Text <> nil then Inc(L, Byte(T^.Text^[0]));
      T := T^.Next;
    end;
  Size.X := L;
end;


destructor TSPanel.Done;
var
  T: PSPanelItem;

  procedure DisposeItems(Item: PSPanelItem);
  var
    T: PSPanelItem;
  begin
    while Item <> nil do
    begin
      T := Item;
      Item := Item^.Next;
      DisposeStr(T^.Text);
      Dispose(T);
    end;
  end;

begin
  DisposeItems(Items);
  inherited Done;
end;


procedure TSPanel.HandleEvent(var Event: TEvent);
var
  Mouse: TPoint;
  T: PSPanelItem;

  function ItemMouseIsIn: PSPanelItem;
  var
    I, K: Word;
    T: PSPanelItem;
  begin
    ItemMouseIsIn := nil;
    if Mouse.Y <> 0 then Exit;
    I := 1;
    T := Items;
    while T <> nil do
    begin
      if T^.Text <> nil then
      begin
        K := I + Byte(T^.Text^[0]);
        if (Mouse.X >= I) and (Mouse.X < K) then
        begin
          ItemMouseIsIn := T;
          Exit;
        end;
        I := K;
      end;
      T := T^.Next;
    end;
  end;

  procedure ExecCommand;
  begin
    if (T <> nil) and CommandEnabled(T^.Command) then
      begin
        Event.What := evCommand;
        Event.Command := T^.Command;
        Event.InfoPtr := nil;
        PutEvent(Event);
        {Owner^.HandleEvent(Event);}
      end;
   {DrawView;
    ClearEvent(Event);}
  end;

begin
  inherited HandleEvent(Event);
  if Event.What = evNothing then
    Exit
  else if Event.What = evMouseMove then
    begin
      repeat
        T := nil;
        MakeLocal(Event.Where, Mouse);
        if T <> ItemMouseIsIn then
          begin
            T := ItemMouseIsIn;
            DrawSelect(T, False);
          end;
        if T = nil then Break;
        if Event.What = evMouseDown then
          begin
            HandleEvent(Event);
            Exit;
          end;
      until not MouseEvent(Event, evMouseMove + evMouseDown);

      if T = nil then DrawView;
      ClearEvent(Event);
    end
  else if Event.What = evMouseDown then
    begin
      T := nil;
      MakeLocal(Event.Where, Mouse);
      repeat
        MakeLocal(Event.Where, Mouse);
        if T <> ItemMouseIsIn then
          begin
            T := ItemMouseIsIn;
            DrawSelect(T, True);
          end;
      until not MouseEvent(Event, evMouseMove);
      ExecCommand;
      {DrawSelect(T, False);}
      DrawView;
      ClearEvent(Event);
    end
  else if Event.What = evBroadCast then
    if Event.Command = cmCommandSetChanged then DrawView;
end;


procedure TSPanel.DrawSelect(Selected: PSPanelItem; Pressed: Boolean);
var
  B: TDrawBuffer;
  CIndex: Byte;
  Colors: array[0..7] of Byte;
  T: PSPanelItem;
  i: Integer;

begin
  for i := 0 to 7 do
    case i of
      0: Colors[0] := $30;
      1: Colors[1] := $37;
      2: Colors[2] := Colors[0];
      3: Colors[3] := Colors[1];
      4: Colors[4] := $3E;
      5: Colors[5] := $37;
      6: Colors[6] := $2E;
      7: Colors[7] := $27;
    end;
   {case i of
      2, 3: Colors[i] := Colors[i-2]
      else
         Colors[i] := GetColor[i];
    end;}

  MoveChar(B, ' ', Colors[0], Size.X);
  MoveChar(B, '[', Colors[0], 1);
  i := 1;
  T := Items;
  while T <> nil do
    begin
      if T^.Text <> nil then
        begin
          if Byte(T^.Text^[0]) + i < Size.X then
            begin
              { CIndex  00000000          }
              {               Disabled }
              {               Presssed }
              {               Selected }
              {              000 - 0      }
              {              001 - 1      }
              {              010 - 2 <- 0 }
              {              011 - 3 <- 1 }
              {              100 - 4      }
              {              101 - 5      }
              {              110 - 6      }
              {              111 - 7      }
              CIndex := Byte(Not(CommandEnabled(T^.Command))) +
                (Byte(Pressed) shl 1) + ((Byte(T = Selected)) shl 2);
              {
              if CommandEnabled(T^.Command) then
                if T = Selected then
                  Color := CSel
                else
                  Color := CNormal
              else if T = Selected then
                  Color := CSelDisabled
                else
                  Color := CDisabled;
              }
              MoveStr(B[i], T^.Text^, Colors[CIndex]);
            end;
          Inc(i, Byte(T^.Text^[0]));
        end;
      T := T^.Next;
    end;
  MoveChar(B[i], ']', Colors[0], 1);
  WriteLine(0, 0, Size.X, 1, B);
end; { DrawSelect }


procedure TSPanel.Draw;
begin
  DrawSelect(nil, false);
end;

{ NewSPanelItem }

function NewSPanelItem(AText: String; ACommand: Word; ANext: PSPanelItem): PSPanelItem;
var
  T: PSPanelItem;
begin
  New(T);
  with T^ do
    begin
      Text := NewStr(AText);
      Command := ACommand;
      Next := ANext;
    end;
  NewSPanelItem := T;
end;


end.