{*******************************************************}
{                                                       }
{       Turbo Pascal 7.0                                }
{       FreePacal                                       }
{       Delphi 4                                        }
{       TV index object                                 }
{                                                       }
{       Copyright (c) 1995-99 by Alexs                  }
{                                                       }
{*******************************************************}

{$IFNDEF FPC}
{$O+}
{$ENDIF}
{$I Define.inc}
unit TVI_Unit;
interface
uses Objects, Ext_Str, DAE, Dos, DBFTable;
const
  defTVIExt:string[4]='.ITV';
type
  PTVIndexHeader = ^TTVIndexHeader;
  TTVIndexHeader = record
    FileOpen:boolean;
    Duplicate:boolean;
    TypeKey:TFieldType;
    IndexsExspresen:string;
  end;

type
  PTVIndexRecord = ^TTVIndexRecord;
  TTVIndexRecord = record
    RecordPos:longint;
    Str:PString;
  end;

type
  TIndexFunc=function(Key1,Key2:Pointer):longint;
type
  PIndexsCollection = ^TIndexsCollection;
  TIndexsCollection = object(TSortedCollection)
    Name:PString;
    Stream:PStream;
    TVIndexHeader:PTVIndexHeader;
    CountInUse:integer;
    constructor Init;
    destructor Done;virtual;
    constructor Load(var S:TStream);
    function Compare(Key1, Key2: Pointer): Integer; virtual;
    procedure FreeItem(Item: Pointer); virtual;
    function GetItem(var S: TStream): Pointer; virtual;
    procedure PutItem(var S: TStream; Item: Pointer); virtual;
  end;

  PTVIndex = ^TTVIndex;
  TTVIndex = object(TIndexObject)
     Modifi:boolean;
     TVIndexHeader:PTVIndexHeader;
     Keys:PIndexsCollection;
     constructor Init(ALink:PDataSet; AFileName:FNameStr);
     constructor Create(ALink:PDataSet; AFileName:FNameStr; AField:string;
                        AOptions:TIndexOptions);
     destructor Done;virtual;
     procedure ReadHeader;
     procedure WriteHeader;
     procedure Flush;
     function FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;virtual;
     function IndexExpresion:string;virtual;
     function NextKey:Longint;virtual;
     function PrevKey:Longint;virtual;
     function Unique:Boolean;virtual;
     procedure AddKey( AKey:String; ALink:LongInt);virtual;
     procedure DeleteKey( AKey:String; ALink:longint);virtual;
     function FirstKey:Longint;virtual;
     function LastKey:Longint;virtual;
     procedure DeleteCyrrent;virtual;
     procedure EmptyIndex; virtual;
  end;

const
  RIndexsCollection: TStreamRec = (
     ObjType: 3200;
     VmtLink: Ofs(TypeOf(TIndexsCollection)^);
     Load:    @TIndexsCollection.Load;
     Store:   @TIndexsCollection.Store
  );
function TVIOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet):PIndexObject;far;
function TVICreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;far;
type
  TDoReIndex=procedure(All,Pos:Longint);
implementation
uses Ext_Dos;
const
   DataCollection:PCollection = nil;

  { TIndexsCollection }

function GetIndexCollection(Name:string):PIndexsCollection;
var
  P:PIndexsCollection;
function Test(P:PIndexsCollection):boolean;far;
begin
  Test:=Name=stUpCase(P^.Name^);
end;
begin
  Name:=stUpCase(Name);
  P:=DataCollection^.FirstThat(@Test);
  if P<>nil then Inc(P^.CountInUse);
  GetIndexCollection:=P;
end;

procedure ReturnIndexCollection(P:PIndexsCollection);
begin
  if DataCollection^.IndexOf(P)=-1 then exit;
  Dec(P^.CountInUse);
  if P^.CountInUse=0 then Dispose(P, Done);
end;


constructor TIndexsCollection.Init;
begin
  inherited Init(500,500);
  DataCollection^.Insert(@Self);
  CountInUse:=1;
end;

constructor TIndexsCollection.Load(var S:TStream);
begin
  inherited Load(S);
  DataCollection^.Insert(@Self);
  CountInUse:=1;
end;

destructor TIndexsCollection.Done;
begin
  DataCollection^.Delete(@Self);
  Stream^.Seek(0);
  Stream^.Write(TVIndexHeader^,SizeOf(TTVIndexHeader));
  Stream^.Put(@Self);
  Dispose(Stream, Done);
  DisposeStr(Name);
  FreeMem(TVIndexHeader,SizeOf(TTVIndexHeader));
  inherited Done;
end;

function TIndexsCollection.Compare(Key1, Key2: Pointer): Integer;
var
  AKey1:PTVIndexRecord absolute Key1;
  AKey2:PTVIndexRecord absolute Key2;
  C:Longint;
begin
  C:=ASCIICompare(AKey1^.Str^,AKey2^.Str^);
  if C<>0 then Compare:=C
  else
  if (Duplicates) and (AKey1^.RecordPos<>-1) and (AKey2^.RecordPos<>-1) then
  begin
     C:=AKey1^.RecordPos-AKey2^.RecordPos;
     if C>0 then Compare:=1
     else if C<0 then Compare:=-1
     else Compare:=0;
  end
  else Compare:=0;
end;

procedure TIndexsCollection.FreeItem(Item: Pointer);
var P:PTVIndexRecord absolute Item;
begin
  DisposeStr(P^.Str);
  Dispose(P);
end;

function TIndexsCollection.GetItem(var S: TStream): Pointer;
var P:PTVIndexRecord;
begin
  New(P);
  S.Read(P^.RecordPos, SizeOf(Longint));
  P^.Str:=S.ReadStr;
  GetItem:=P;
end;

procedure TIndexsCollection.PutItem(var S: TStream; Item: Pointer);
var P:PTVIndexRecord absolute Item;
begin
  S.Write(P^.RecordPos, SizeOf(Longint));
  S.WriteStr(P^.Str);
end;

{TTVIndex}

constructor TTVIndex.Init(ALink:PDataSet; AFileName:FNameStr);
var
  AOptions:TIndexOptions;
begin
   TObject.Init;
   AFileName:=ReplaceExt(AFileName, defTVIExt, false);
   Owner:=ALink;
   Keys:=GetIndexCollection(AFileName);
   if Keys<>nil then
   begin
     TVIndexHeader:=Keys^.TVIndexHeader;
     PStream(IndexStream):=Keys^.Stream;
   end
   else
   begin
     PStream(IndexStream):=New(PDosStream, Init(AFileName, stOpen));
     GetMem(TVIndexHeader, SizeOf(TTVIndexHeader));
     FillChar(TVIndexHeader^, SizeOf(TTVIndexHeader), 0);
     ReadHeader;
     Keys:=PIndexsCollection(IndexStream^.Get);
     Keys^.TVIndexHeader:=TVIndexHeader;
     Keys^.Stream:=IndexStream;
     Keys^.Name:=NewStr(AFileName);
   end;
   FileName:=Keys^.Name;
   IndexField:=GetIndexsField(Owner, IndexExpresion, AOptions);
   Options:=Options+AOptions;
   if IndexField=nil then
   begin
     ReturnIndexCollection(Keys);
     DisposeStr(FileName);
     fail;
   end;
   if Unique then Options:=Options+[ixUnique];
end;

constructor TTVIndex.Create(ALink:PDataSet; AFileName:FNameStr;
                            AField:string; AOptions:TIndexOptions);
begin
   TObject.Init;
   AFileName:=ReplaceExt(AFileName, defTVIExt, false);
   Owner:=ALink;
   GetMem(TVIndexHeader, SizeOf(TTVIndexHeader));
   FillChar(TVIndexHeader^, SizeOf(TTVIndexHeader), 0);
   Options:=AOptions;
   TVIndexHeader^.IndexsExspresen:=AField;
   TVIndexHeader^.Duplicate:=not (ixUnique in Options);
   Keys:=GetIndexCollection(AFileName);
   IndexField:=GetIndexsField(Owner, IndexExpresion, AOptions);
   if (Keys<>nil) or (IndexField=nil) then
   begin
     if Keys<>nil then DBError(edbFileAlreadyOpen, AFileName);
     ReturnIndexCollection(Keys);
     DisposeStr(FileName);
     fail;
   end;
   Options:=Options+AOptions;
   PStream(IndexStream):=New(PDosStream, Init(AFileName, stCreate));
   Keys:=New(PIndexsCollection,Init);
   Keys^.Duplicates:=TVIndexHeader^.Duplicate;
   Keys^.Stream:=IndexStream;
   Keys^.TVIndexHeader:=TVIndexHeader;
   Keys^.Name:=NewStr(AFileName);
   FileName:=Keys^.Name;
   WriteHeader;
   Reindexs;
end;

destructor TTVIndex.Done;
begin
  ReturnIndexCollection(Keys);
  TObject.Done;
end;

procedure TTVIndex.ReadHeader;
begin
  IndexStream^.Seek(0);
  IndexStream^.Read(TVIndexHeader^,SizeOf(TTVIndexHeader));
end;

procedure TTVIndex.WriteHeader;
begin
  IndexStream^.Seek(0);
  IndexStream^.Write(TVIndexHeader^,SizeOf(TTVIndexHeader));
end;

procedure TTVIndex.Flush;
begin
  WriteHeader;
  IndexStream^.Seek(SizeOf(TTVIndexHeader));
  IndexStream^.Put(Keys);
end;

function TTVIndex.FindKey(Key:string; PtrRec:longint; SoftFind:Boolean):LongInt;
var P:TTVIndexRecord;
    B:Boolean;
    S:string;
begin
  ClearKey;
  P.Str:=@Key;
  P.RecordPos:=PtrRec;
  B:=Keys^.Search(@P,PP);
  if B or (SoftFind and (PP<Keys^.Count)) then
  begin
     if (PtrRec=MaxLongint) and (PP>0) then Dec(PP);
     FindKey:=PTVIndexRecord(Keys^.At(PP))^.RecordPos
  end
  else FindKey:=-1;
  if (PP>=Keys^.Count) or EndRange(PTVIndexRecord(Keys^.At(PP))^.Str^) then iStatus:=iStatus+[iBofFind];
  if (PP>=Keys^.Count) or StartRange(PTVIndexRecord(Keys^.At(PP))^.Str^) then
     iStatus:=iStatus+[iEofFind];
end;

function TTVIndex.IndexExpresion:string;
begin
  IndexExpresion:=TVIndexHeader^.IndexsExspresen;
end;

function TTVIndex.NextKey:Longint;
var
  P:Pointer;
  L:Longint;
  Rec:TTVIndexRecord;
  S:String;
begin
  if (PP<Keys^.Count-1) and
     not EndRange(PTVIndexRecord(Keys^.At(PP+1))^.Str^) then
  begin
    Inc(PP);
    iStatus:=iStatus-[iBofFind];
  end
  else iStatus:=iStatus+[iEofFind];
  if (PP>=Keys^.Count) or EndRange(PTVIndexRecord(Keys^.At(PP))^.Str^) then
     iStatus:=iStatus+[iBofFind];
  if (PP>=Keys^.Count) or StartRange(PTVIndexRecord(Keys^.At(PP))^.Str^) then
     iStatus:=iStatus+[iEofFind];
  if not (iEofFind in iStatus) then NextKey:=PTVIndexRecord(Keys^.At(PP))^.RecordPos
  else NextKey:=-1;
end;

function TTVIndex.PrevKey:Longint;
var
  P:Pointer;
  L:Longint;
  Rec:TTVIndexRecord;
  S:String;
begin
  OldPP:=PP;
  if PP>0 then
  begin
    Dec(PP);
    if StartRange(PTVIndexRecord(Keys^.At(PP))^.Str^) then
       iStatus:=iStatus+[iBofFind]
    else iStatus:=iStatus-[iEofFind];
  end
  else iStatus:=iStatus+[iBofFind];
  if (PP=-1) or (iBofFind in iStatus) then
  begin
    PrevKey:=-1;
    PP:=OldPP;
  end
  else PrevKey:=PTVIndexRecord(Keys^.At(PP))^.RecordPos;
end;

function TTVIndex.Unique:Boolean;
begin
  Unique:=not TVIndexHeader^.Duplicate;
end;

procedure TTVIndex.AddKey( AKey:String; ALink:LongInt);
var P:PTVIndexRecord;
begin
  New(P);
  P^.RecordPos:=ALink;
  P^.Str:=NewStr(AKey);
  Keys^.Insert(P);
  Keys^.Search(P,PP);
  if PP>=Keys^.Count then PP:=Keys^.Count-1;
end;

procedure TTVIndex.DeleteKey( AKey:String; ALink:longint);
var P:TTVIndexRecord;
    I:integer;
begin
  P.Str:=@AKey;
  P.RecordPos:=ALink;
  if Keys^.Search(@P,I) then Keys^.AtFree(i);
  if i<Keys^.Count then PP:=i
  else PP:=Keys^.Count;
end;


function TTVIndex.FirstKey:Longint;
var S:String;
begin
  ClearKey;
  S:=StartRangeValue;
  if S<>'' then FirstKey:=FindKey(S, 0, true)
  else
  begin
    PP:=0;
    FirstKey:=PTVIndexRecord(Keys^.At(PP))^.RecordPos;
  end;
end;

function TTVIndex.LastKey:Longint;
var S:String;
begin
  ClearKey;
  S:=EndRangeValue;
  if S<>'' then LastKey:=FindKey(S, MaxLongint, true)
  else
  begin
    PP:=Keys^.Count-1;
    LastKey:=PTVIndexRecord(Keys^.At(PP))^.RecordPos
  end;
end;

procedure TTVIndex.DeleteCyrrent;
var P:TTVIndexRecord;
    P_P:PTVIndexRecord;
    I:integer;
    S:String;

function Test(Item:PTVIndexRecord):boolean;
begin
  Test:=Keys^.Compare(Item, @P)=0;
end;

begin
  S:=IndexField^.IndexString;
  P.Str:=@S;
  P.RecordPos:=Owner^.RecNo;
  if Keys^.Search(@P, i) then Keys^.AtFree(i);
  FirstKey;
end;

procedure TTVIndex.EmptyIndex;
begin
  PP:=0;
  Keys^.FreeAll;
end;

procedure RegisterTVI_Unit;
begin
  RegisterType(RIndexsCollection);
end;

var
  OldExit:TDBExitProc;
  OldBegin:TDBExitProc;
procedure InitDataSets;far;
begin
  if DataCollection = nil then
     DataCollection:=New(PCollection, Init( 10 ,5));
  OldBegin;
  RegisterTVI_Unit;
end;

procedure DoneDataSets;far;
begin
  OldExit;
  Dispose( DataCollection, Done);
  DataCollection:=nil;
end;

const
   OldOnOpenIndex:TIndexOpen=nil;
   OldOnCreateIndex:TIndexCreate=nil;

function TVIOpen(FileName:FNameStr; var Ext:ExtStr; DS:PDataSet):PIndexObject;
begin
  if Ext=defTVIExt then
     TVIOpen:=New(PTVIndex, Init(DS, FileName))
  else
  if Assigned(OldOnOpenIndex) then
     TVIOpen:=OldOnOpenIndex(FileName,Ext,DS)
  else TVIOpen:=nil;
end;

function TVICreate( FileName:FNameStr; var Ext:ExtStr; DS:PDataSet;
                    AField:string; AOptions:TIndexOptions):PIndexObject;
begin
  if Ext=defTVIExt then
     TVICreate:=New(PTVIndex, Create(DS, FileName, AField, AOptions))
  else
  if Assigned(OldOnOpenIndex) then
     TVICreate:=OldOnOpenIndex(FileName,Ext, DS)
  else TVICreate:=nil;
end;
begin
   OldOnOpenIndex:=OnOpenIndex;
   OldOnCreateIndex:=OnCreateIndex;
   OnCreateIndex:=TVICreate;
   OnOpenIndex:=TVIOpen;

   OldExit:=DBExitProc;
   DBExitProc:=DoneDataSets;
   OldBegin:=DBBeginProc;
   DBBeginProc:=InitDataSets;
end.
