{************************************************}
{                                                }
{   Extended crt unit                            }
{                                                }
{   Turbo Pascal 7                               }
{   FreePacal                                    }
{   Delphi 2 - 5                                 }
{   Virtual Pascal 2.1                           }
{                                                }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}

{$I Define.inc}

{$IFDEF VER70}
{$O+}
{$ENDIF}

unit Ext_Crt;

interface
uses
{$IFDEF WIN32}
  Windows,
{$ENDIF}
{$IFDEF VER70}
  VMM,
{$ENDIF}
  Objects;

const
{ CRT modes }
  BW40          = 0;            { 40x25 B/W on Color Adapter }
  CO40          = 1;            { 40x25 Color on Color Adapter }
  BW80          = 2;            { 80x25 B/W on Color Adapter }
  CO80          = 3;            { 80x25 Color on Color Adapter }
  Mono          = 7;            { 80x25 on Monochrome Adapter }
  Font8x8       = 256;          { Add-in for ROM font }
{ Mode constants for 3.0 compatibility }
  C40           = CO40;
  C80           = CO80;
{ Foreground and background color constants }
  Black         = 0;
  Blue          = 1;
  Green         = 2;
  Cyan          = 3;
  Red           = 4;
  Magenta       = 5;
  Brown         = 6;
  LightGray     = 7;
{ Foreground color constants }
  DarkGray      = 8;
  LightBlue     = 9;
  LightGreen    = 10;
  LightCyan     = 11;
  LightRed      = 12;
  LightMagenta  = 13;
  Yellow        = 14;
  White         = 15;
{ Add-in for blinking }
  Blink         = 128;

  { Possible cursor types }
  crHidden        = 0;
  crUnderLine     = 1;
  crBlock         = 2;
  crHalfBlock     = 3;

type
{$IFNDEF WIN32}
  TCharInfo = record
    Char:Char;
    Attr:byte;
  end;
{$ENDIF}
  PScreenBuffer = ^TScreenBuffer;
{$IFDEF WIN32}
  TScreenBuffer = array [0..(1024*32) - 1] of TCharInfo;
{$ELSE}
  TScreenBuffer = array[1..25, 1..80] of TCharInfo;
{$ENDIF}

type
  TSerPacket = record
    Info:Word;
    SerNo: Longint;
    Vol: Array[1..11] of char;
    FileType: Array[1..8] of char;
  end;

procedure SaveScreen;
procedure RestoreScreen;
function WaitPress:word;
procedure GotoXY(X,Y:byte);
procedure ClrScr;
procedure Delay(Wait:LongInt);
procedure Sleep(Wait:LongInt); {in 1/1 000 000 sec}
function ReadKey:char;
function ReadKeyExt:word;
function WhereX:byte;
function WhereY:byte;
procedure OutTextXY(X,Y:integer; Color:byte; Text:string);
procedure SetTextAttr(Color:word);
procedure TextBackGround(Color:byte);
procedure TextColor(Color:byte);
function Keypressed:boolean;
procedure OutBuffer(X,Y,W:integer;var Buf);
function ReadScreenChar(X,Y:integer):char;
function ReadScreenAttr(X,Y:integer):byte;
procedure WriteScreenChar(X,Y:integer; AChar:char);
procedure WriteScreenAttr(X,Y:integer; Attr:byte);
{ Ext CRT routins}
procedure SaveScreenVMM;
procedure RestoreScreenVMM;
{$IFDEF WIN32}
function GetCursorType: Word; { Return the cursor type: Hidden, UnderLine or Block }
procedure SetCursorType(NewType: Word); { Set the cursor to the given type }
{$ENDIF}

const
{$IFNDEF FPC}
  Screen:PScreenBuffer = nil;
{$ELSE}
  Screen:^TCharInfo = nil;
{$ENDIF}
var
  OldX,OldY:Byte;
  TextAttr:word;
{$IFDEF WIN32}
  hStdout,hStdin:THandle;
  ConsoleInfo:TConsoleScreenBufferInfo;
{$ENDIF}

implementation
{$IFDEF VP}
uses Drivers;
{$ENDIF}
{$IFDEF WIN32}
{$I Ext_Crt.w32}
{$ENDIF}
{$IFDEF OS_DOS}
{$I Ext_Crt.dos}
{$ENDIF}

procedure SaveScreen;
begin
  if Screen=nil then GetMem(Screen, SizeOf(TScreenBuffer));
  InternalMoveFromScreen;
  OldX:=WhereX;
  OldY:=WhereY;
end;

procedure RestoreScreen;
begin
  if Screen = nil then exit;
  InternalMoveToScreen;
  GotoXY(OldX,OldY);
end;

{$IFDEF VER70}
function WaitPress:word;assembler;
asm
@@1: { CMP ButtonCount,false
      JE @@2
      MOV AX,3
      INT 33
      XOR AX,AX
      OR BL,BL
      JNE @@3
@@2:}  MOV AH,1
      INT 16h
      JE @@1
      MOV AH,0
      INT 16h
      OR AX,AX
      JE @@1
@@3:
end;
{$ELSE}
function WaitPress:word;
begin
  ReadKey;
end;
{$ENDIF}


{$IFDEF VER70}
procedure OutBuffer(X,Y,W:integer;var Buf);
begin
end;

function WhereX:byte;assembler;
asm
        mov AH,03h
        xor BH,BH
        int 10h
        mov AL,DL
end;

function WhereY:byte;assembler;
asm
        mov AH,03h
        xor BH,BH
        int 10h
        mov AL,DH
end;


function ReadKey;assembler;
asm
   MOV  AX,0
   INT  16h
end;

function Keypressed:boolean;
begin
end;

function ReadKeyExt:word;assembler;
asm
   MOV  AX,0
   INT  16h
end;

procedure OutTextXY(X,Y:integer; Color:byte; Text:string);assembler;
asm
        PUSH    DS
        MOV     AX,Y
        MOV     BX,80*2
        MUL     BX
        MOV     BX,X
        SHL     BX,1
        MOV     DI,AX
        ADD     DI,BX
        MOV     AX,SEGB800
        MOV     ES,AX
        XOR     AX,AX
        CLD
        LDS     SI,Text
        LODSB
        MOV     CX,AX
        JCXZ    @@2
        MOV     AH,Color
@@L1:
        LODSB
        STOSW
        LOOP    @@L1
@@2:
        POP     DS
end;

procedure Delay(Wait:LongInt);
begin
  Sleep(Wait*1000);
end;

procedure SetTextAttr(Color:word);
begin
end;
procedure TextBackGround(Color:byte);
begin
end;
procedure TextColor(Color:byte);
begin
end;
{.$ELSE}
{$ENDIF}

{$IFDEF WIN32}

{$ENDIF}


{$IFDEF WIN32}
procedure OutTextXY(X,Y:integer; Color:byte; Text:string);
begin
  GotoXY(X,Y);
  SetTextAttr(Color);
  Write(Text);
end;

function Keypressed:boolean;
var
  inBuf:TInputRecord;
  res:DWORD;
begin
  Result:=False;
  PeekConsoleInput(hStdin, inBuf, 1, res);
  if (inBuf.EventType=KEY_EVENT) then
{$IFDEF FPC}
    Result:=inBuf.Event.KeyEvent.bKeyDown;
{$ELSE}
    Result:=inBuf.KeyEvent.bKeyDown;
{$ENDIF}
end;

procedure TextBackGround(Color:byte);
begin
  TextAttr:=TextAttr and $71 or Color;
  SetTextAttr(TextAttr);
end;

procedure TextColor(Color:byte);
begin
  TextAttr:=TextAttr and $F0 or Color;
  SetTextAttr(TextAttr);
end;

procedure SetTextAttr(Color:word);
begin
  TextAttr:=Color;
  SetConsoleTextAttribute(hStdout,Color);
end;

function WhereX:byte;
begin
//  GetConsoleScreenBufferInfo(hStdOut,ConsoleInfo);
//  Result:=ConsoleInfo.dwCursorPosition.X;
end;

function WhereY:byte;
begin
//  GetConsoleScreenBufferInfo(hStdOut,ConsoleInfo);
//  Result:=ConsoleInfo.dwCursorPosition.Y;
end;



type
  TDrawBuffer = array [0..127] of word;
  TConsoleArray = array [1..127] of TCharInfo;
  TWordRec = record
    Lo, Hi: Byte;
  end;

procedure OutBuffer(X,Y,W:integer;var Buf);
(*var
  dwBufferSize, dwBufferCoord: TCoord;
  lpWriteRegion: TSmallRect;
  Buffer:TDrawBuffer absolute Buf;
  Console:TConsoleArray;
  i:word; *)
begin
//  dwBufferSize:=ConsoleInfo.dwSize;
(*  dwBufferSize.X:=80;
  dwBufferSize.Y:=1;
  dwBufferCoord.X:=0;
  dwBufferCoord.Y:=0;
  lpWriteRegion.Left:=X;
  lpWriteRegion.Top:=Y;
  lpWriteRegion.Right:=W+X-1;
  lpWriteRegion.Bottom:=Y;
  for i := 0 to W do
  begin
    Console[i].AsciiChar := Char(TWordRec(Buffer[i-1]).Lo);
    Console[i].Attributes := TWordRec(Buffer[i-1]).Hi;
  end;
  {$IFDEF FPC}
  WriteConsoleOutput(hStdout, PCharInfo(@Console)^, dwBufferSize, dwBufferCoord,  @lpWriteRegion);
  {$ELSE}
  WriteConsoleOutput(hStdout, @Console, dwBufferSize, dwBufferCoord,  lpWriteRegion);
  {$ENDIF} *)
end;
{$ENDIF}

{ CRT }
procedure SaveScreenVMM;
{$IFDEF VER70}
var Ptr:pointer;
    Page:PVMMPage;
begin
  Ptr:=GetVMMPage(TPtrHandle(Screen), 4000, Page);
  Move( Mem[VideoMemory:0], Ptr^, 4000);
  OldX:=WhereX;
  OldY:=WhereY;
end;
{$ELSE}
begin
end;
{$ENDIF}

procedure RestoreScreenVMM;
{$IFDEF VER70}
var Ptr:pointer;
    Page:PVMMPage;
begin
  Ptr:=UpdateVMMPage(TPtrHandle(Screen), Page);
  Move( Ptr^, Mem[VideoMemory:0],4000);
  GotoXY(OldX,OldY);
end;
{$ELSE}
begin
end;
{$ENDIF}

{$IFDEF WIN32}
function GetCursorType: Word;
var ConsoleCursorInfo : TConsoleCursorInfo;
begin
  GetConsoleCursorInfo(hStdOut,ConsoleCursorInfo);
  if not ConsoleCursorInfo.bvisible then GetCursorType:=crHidden
  else
  case ConsoleCursorInfo.dwSize of
    1..30:GetCursorType:=crUnderline;
    31..70:GetCursorType:=crHalfBlock;
    71..100:GetCursorType:=crBlock;
  end;
end;


procedure SetCursorType(NewType: Word);
var ConsoleCursorInfo : TConsoleCursorInfo;
begin
  GetConsoleCursorInfo(hStdOut,ConsoleCursorInfo);
  if newType=crHidden then ConsoleCursorInfo.bvisible:=false
  else
  begin
    ConsoleCursorInfo.bvisible:=true;
    case NewType of
      crUnderline:ConsoleCursorInfo.dwSize:=20;
      crHalfBlock:ConsoleCursorInfo.dwSize:=50;
      crBlock:ConsoleCursorInfo.dwSize:=100;
    end
  end;
  SetConsoleCursorInfo(hStdOut,ConsoleCursorInfo);
end;
{$ENDIF}

{$IFDEF WIN32}
initialization
  hStdout:=GetStdHandle(STD_OUTPUT_HANDLE);
  hStdin:=GetStdhandle(STD_INPUT_HANDLE);
  FlushConsoleInputBuffer(hStdin);
  GetConsoleScreenBufferInfo(hStdOut,ConsoleInfo);
  GotoXY(ConsoleInfo.dwCursorPosition.X,ConsoleInfo.dwCursorPosition.Y);
finalization
  if Screen<>nil then FreeMem(Screen, SizeOf(TScreenBuffer));
{$ENDIF WIN32}
end.
