{************************************************}
{                                                }
{   Turbo Pascal 7.0                             }
{   FreePacal                                    }
{   Delphi 4                                     }
{   Data access unit                             }
{   Copyright (c) 1995-2000 by Alexs             }
{                                                }
{************************************************}
{$IFDEF VER70}
{$O+}
{$ENDIF}
{$I Define.inc}

unit Parser;

interface
uses Objects;

type
  TParserType = (ptEmpty, ptString, ptNumeric, ptBoolean);
  TParserVal = record
    case fType:TParserType of
     ptEmpty:();
     ptString:(StrVal:PString);
     ptNumeric:(NumVal:Double);
     ptBoolean:(BoolVal:Boolean)
  end;

  PParserValue = ^TParserValue;
  TParserValue = object(TObject)
    Value:TParserVal;
    destructor Done;virtual;
    procedure Clear;
    procedure AssignStr(AValue:string);
    procedure AssignNum(AValue:Double);
    procedure AssignBool(AValue:boolean);
    function AsString:string;
    function AsNumeric:Double;
    function AsBoolean:Boolean;
    procedure Assign(P:PParserValue);
    procedure AddVal(var P1,P2:TParserValue);
    procedure SubVal(var P1,P2:TParserValue);
    procedure MulVal(var P1,P2:TParserValue);
    procedure DivVal(var P1,P2:TParserValue);
    procedure GtVal(var P1,P2:TParserValue);
    procedure LtVal(var P1,P2:TParserValue);
    procedure EqVal(var P1,P2:TParserValue);
    procedure NEVal(var P1,P2:TParserValue);
    procedure GEVal(var P1,P2:TParserValue);
    procedure LEVal(var P1,P2:TParserValue);
  end;

type
  PParser = ^TParser;
  TGetPValueEvent = procedure(Owner:PObject; Sender:PParser; const s:String; var v:TParserValue);
  TFunctionEvent = procedure(Owner:PObject; Sender:PParser; const name:String; var p1,p2,p3:string;var Val:TParserValue);
  TParser = object(TObject)
  private
    function GetIdentify(const s:String; var i:Integer):String;
    function GetString(const s:String; var i:Integer):String;
    procedure Get3Parameters(const s:String; var i:Integer; var s1,s2,s3:String);
  public
    OnGetValue: TGetPValueEvent;
    OnFunction: TFunctionEvent;
    Owner:PObject;
    Format: (fNone,fDate,fTime);
    constructor Init(AOwner:PObject);
    function Str2OPZ(s:String):String;
    procedure CalcOPZ(const s:String;var Result:TParserValue);
    procedure Calc(const s:String; var Result:TParserValue);
  end;

implementation
uses Ext_Sys, Ext_Math, Ext_Str, Ext_Dos;

  {TParserValue}
destructor TParserValue.Done;
begin
  Clear;
  inherited Done;
end;

procedure TParserValue.Clear;
begin
  if Value.fType=ptString then DisposeStr(Value.StrVal);
  Value.fType:=ptEmpty;
end;

procedure TParserValue.AssignStr(AValue:string);
begin
  Clear;
  Value.fType:=ptString;
  Value.StrVal:=NewStr(AValue)
end;

procedure TParserValue.AssignNum(AValue:Double);
begin
  Clear;
  Value.fType:=ptNumeric;
  Value.NumVal:=AValue
end;

function TParserValue.AsString:string;
begin
  case Value.fType of
    ptString:if (Value.StrVal<>nil) then AsString:=Value.StrVal^ else AsString:='';
    ptNumeric:AsString:=FloatToStr(AsNumeric);
    ptBoolean:AsString:=BooleanToStr(AsBoolean);
  else
    AsString:='';
  end;
end;

function TParserValue.AsNumeric:Double;
begin
  if Value.fType=ptNumeric then AsNumeric:=Value.NumVal else
  if Value.fType=ptBoolean then AsNumeric:=ord(AsBoolean)
  else AsNumeric:=0;
end;

function TParserValue.AsBoolean:Boolean;
begin
  if Value.fType=ptBoolean then AsBoolean:=Value.BoolVal else
  if Value.fType=ptNumeric then AsBoolean:=AsNumeric<>0
  else AsBoolean:=false;
end;

procedure TParserValue.AssignBool(AValue:boolean);
begin
  Clear;
  Value.fType:=ptBoolean;
  Value.BoolVal:=AValue;
end;

procedure TParserValue.Assign(P:PParserValue);
begin
  Clear;
  if P<>nil then
  begin
    Value.fType:=P^.Value.fType;
    case Value.fType of
      ptString:Value.StrVal:=NewStr(DinamicStr(P^.Value.StrVal));
      ptNumeric:Value.NumVal:=P^.Value.NumVal;
      ptBoolean:Value.BoolVal:=P^.Value.BoolVal;
    end;
  end;
end;

procedure TParserValue.AddVal(var P1,P2:TParserValue);
begin
  if @P1<>@Self then Assign(@P1);
  if Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:Ext_Str.AssignStr(Value.StrVal,DinamicStr(Value.StrVal)+DinamicStr(P2.Value.StrVal));
    ptNumeric:Value.NumVal:=Value.NumVal+P2.Value.NumVal;
    ptBoolean:Value.BoolVal:=Value.BoolVal or P2.Value.BoolVal;
  else
   {error}
  end;
end;

procedure TParserValue.SubVal(var P1,P2:TParserValue);
begin
  if @P1<>@Self then Assign(@P1);
  if Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptNumeric:Value.NumVal:=Value.NumVal-P2.Value.NumVal;
  else
   {error}
  end;
end;

procedure TParserValue.MulVal(var P1,P2:TParserValue);
begin
  if @P1<>@Self then Assign(@P1);
  if Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptNumeric:Value.NumVal:=Value.NumVal*P2.Value.NumVal;
    ptBoolean:Value.BoolVal:=Value.BoolVal and P2.Value.BoolVal;
  else
   {error}
  end;
end;

procedure TParserValue.DivVal(var P1,P2:TParserValue);
begin
  if @P1<>@Self then Assign(@P1);
  if Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptNumeric:Value.NumVal:=Value.NumVal / P2.Value.NumVal;
  else
   {error}
  end;
end;

procedure TParserValue.GtVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)>DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal>P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal>P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

procedure TParserValue.LtVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)<DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal<P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal<P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

procedure TParserValue.EqVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)=DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal=P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal=P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

procedure TParserValue.GEVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)>=DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal>=P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal>=P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

procedure TParserValue.LEVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)<=DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal<=P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal<=P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

procedure TParserValue.NEVal(var P1,P2:TParserValue);
var B:boolean;
begin
  if P1.Value.fType<>P2.Value.fType then
  else
  case Value.fType of
    ptString:B:=DinamicStr(P1.Value.StrVal)<>DinamicStr(P2.Value.StrVal);
    ptNumeric:B:=P1.Value.NumVal<>P2.Value.NumVal;
    ptBoolean:B:=P1.Value.BoolVal<>P2.Value.BoolVal;
  else
   {error}
  end;
  AssignBool(B);
end;

const
  ttGe = #1; ttLe = #2;
  ttNe = #3; ttOr = #4; ttAnd = #5;
  ttInt = #6;  ttFrac = #7;
  ttUnMinus = #9; ttUnPlus = #10; ttStr = #11;
  ttNot = #12; ttMod = #13; ttRound = #14;

function GetBrackedVariable(s:String; var i,j:Integer): String;
var c: Integer;
    fl1,fl2: Boolean;
begin
  j := i; fl1 := True; fl2 := True; c := 0;
  GetBrackedVariable := '';
  if s = '' then Exit;
  Dec(j);
  repeat
    Inc(j);
    if fl1 and fl2 then
      if s[j] = '[' then
      begin
        if c = 0 then i := j;
        Inc(c);
      end
      else if s[j] = ']' then Dec(c);
    if fl1 then
      if s[j] = '"' then fl2 := not fl2;
    if fl2 then
      if s[j] = '''' then fl1 := not fl1;
  until (c = 0) or (j >= Length(s));
  GetBrackedVariable := Copy(s,i+1,j-i-1);
end;

constructor TParser.Init(AOwner:PObject);
begin
  inherited Init;
  Owner:=AOwner;
end;

procedure TParser.CalcOPZ(const s:String;var Result:TParserValue);
var i,j,k,i1,st,ci,cn: Integer;
    s1,s2,s3,s4,s5: String;
    nm: Array[1..8] of TParserValue;
    R:TParserValue;
begin
  for i:=1 to 8 do Nm[i].Init;
  R.Init;
  st := 1;
  i := 1;
  Format := fNone;
  while i <= Length(s) do
  begin
    j := i;
    case s[i] of
      '+',ttOr:Nm[st-2].AddVal(nm[st-2],nm[st-1]);
      '-':Nm[st-2].SubVal(nm[st-2],nm[st-1]);
      '*',ttAnd:Nm[st-2].MulVal(nm[st-2],nm[st-1]);
      '/':Nm[st-2].DivVal(nm[st-2],nm[st-1]);
      '>':Nm[st-2].GTVal(nm[st-2],nm[st-1]);
      '<':Nm[st-2].LTVal(nm[st-2],nm[st-1]);
      '=':Nm[st-2].EQVal(nm[st-2],nm[st-1]);
      ttNe:Nm[st-2].NEVal(nm[st-2],nm[st-1]);
      ttGe:Nm[st-2].GEVal(nm[st-2],nm[st-1]);
      ttLe:Nm[st-2].LEVal(nm[st-2],nm[st-1]);
      ttInt:Nm[st-1].AssignNum(INT(nm[st-1].AsNumeric));
      ttFrac:Nm[st-1].AssignNum(Frac(nm[st-1].AsNumeric));
      ttRound:Nm[st-1].AssignNum(Round(nm[st-1].AsNumeric));
      ttUnMinus:Nm[st-1].AssignNum(-nm[st-1].AsNumeric);
      ttUnPlus:;
      ttStr:Nm[st-1].AssignStr(nm[st-1].AsString);
      ttNot:Nm[st-1].AssignBool(not Nm[st-1].AsBoolean);
      ttMod:Nm[st-2].AssignNum(Round(nm[st-2].AsNumeric) mod
                               Round(nm[st-1].AsNumeric));
      '%':
        begin
          if s[i+1] in ['d','D'] then Format := fDate;
          if s[i+1] in ['t','T'] then Format := fTime;
          Inc(i);
        end;
      ' ': ;
      '[':
        begin
          k := i;
          s1 := GetBrackedVariable(s,k,i);
          if Assigned(OnGetValue) then OnGetValue(Owner, @Self, s1,nm[st]);
          Inc(st);
        end
      else
        begin
          if s[i] = '''' then
          begin
            s1 := GetString(s,i);
            while Pos(''''+'''',s1) <> 0 do
              Delete(s1,Pos(''''+'''',s1),1);
            s1 := Copy(s1,2,Length(s1)-2);
            if Format = fNone then Nm[st].AssignStr(S1)
            else
            if Format = fDate then Nm[st].AssignNum(StrToDate(s1))
            else
            if Format = fTime then Nm[st].AssignNum(StrToTime(s1));
            Format := fNone;
            k := i;
          end
          else
          begin
            k := i;
            s1 := GetIdentify(s,k);
            if s1[1] in ['0'..'9', '.', ','] then
            begin
              for i1 := 1 to Length(s1) do
                if s1[i1] in ['.', ','] then s1[i1] := DecimalSeparator;
              nm[st].AssignNum(StrToFloat(s1));
            end
            else if AnsiCompareText(s1, 'TRUE') = 0 then
              nm[st].AssignBool(true)
            else if AnsiCompareText(s1, 'FALSE') = 0 then
              nm[st].AssignBool(false)
            else if s[k] = '[' then
            begin
              s1 := 'GETARRAY(' + s1 + ', ' + GetBrackedVariable(s, k, i) + ')';
              R.Clear;
              Calc(S1,R);
              nm[st].Assign(@R);
              k := i;
            end
            else if s[k] = '(' then
            begin
              s1 := AnsiUpperCase(s1);
              if s1 = 'COPY' then
              begin
                Get3Parameters(s, k, s2, s3, s4);
                R.Clear;
                Calc(S3,R);
                ci := Trunc(R.AsNumeric);
                R.Clear;
                Calc(S4,R);
                cn :=Trunc(R.AsNumeric);
                R.Clear;
                Calc(S2,R);
                nm[st].AssignStr(Copy(R.AsString, ci, cn));
              end
              else if s1 = 'IF' then
              begin
                Get3Parameters(s, k, s2, s3, s4);
                R.Clear;
                Calc(S2,R);
                if R.AsBoolean then
                  s1 := s3 else
                  s1 := s4;
                R.Clear;
                Calc(S1,R);
                nm[st].Assign(@R);
              end
              else if s1 = 'STRTODATE' then
              begin
                Get3Parameters(s, k, s2, s3, s4);
                R.Clear;
                Calc(S2,R);
                nm[st].AssignNum(StrToDate(R.AsString));
              end
              else if s1 = 'STRTOTIME' then
              begin
                Get3Parameters(s, k, s2, s3, s4);
                R.Clear;
                Calc(S2,R);
                nm[st].AssignNum(StrToTime(R.AsString));
              end
              else
              begin
                Get3Parameters(s, k, s2, s3, s4);
                if Assigned(OnFunction) then
                begin
                   OnFunction(Owner, @Self, s1, s2, s3, s4, nm[st]);
                   {nm[st].AssignStr(S5);}
                end;
              end;
              Dec(k);
            end
            else
              if Assigned(OnGetValue) then
                 OnGetValue(Owner, @Self, AnsiUpperCase(s1), nm[st]);
          end;
          i := k;
          Inc(st);
        end;
    end;
    if s[j] in ['+','-','*','/','>','<','=',ttGe,ttLe,ttNe,ttOr,ttAnd,ttMod] then
      Dec(st);
    Inc(i);
  end;
  Result.Assign(@Nm[1]);
  for i:=1 to 8 do Nm[i].Done;
  R.Done;
end;

function TParser.GetIdentify(const s:String; var i:Integer):String;
var k,n: Integer;
begin
  n := 0;
  while (i <= Length(s)) and (s[i] = ' ') do
    Inc(i);
  k := i; Dec(i);
  repeat
    Inc(i);
    while (i <= Length(s)) and not (s[i] in [' ','+','-','*','/','>','<','=','(',')']) do
    begin
      if s[i] = '"' then Inc(n);
      Inc(i);
    end;
  until n mod 2 = 0;
  GetIdentify:= Copy(s, k, i - k);
end;

function TParser.GetString(const s:String; var i:Integer):String;
var k: Integer;
    f: Boolean;
begin
  k := i; Inc(i);
  repeat
    while (i <= Length(s)) and (s[i] <> '''') do
      Inc(i);
    f := True;
    if (i < Length(s)) and (s[i+1] = '''') then
    begin
      f := False;
      Inc(i,2);
    end;
  until f;
  GetString := Copy(s, k, i-k+1);
  Inc(i);
end;

procedure TParser.Get3Parameters(const s:String; var i:Integer; var s1,s2,s3:String);
var c,d,oi,ci: Integer;
begin
  s1 := ''; s2 := ''; s3 := '';
  c := 1; d := 1; oi := i+1; ci := 1;
  repeat
    Inc(i);
    if s[i] = '(' then Inc(c)
    else if s[i] = ')' then Dec(c);
    if s[i] = '''' then
      if d = 1 then Inc(d) else d := 1;
    if (s[i] = ',') and (c = 1) and (d = 1) then
    begin
      if ci = 1 then
        s1 := Copy(s,oi,i-oi)
      else
        s2 := Copy(s,oi,i-oi);
      oi := i+1; Inc(ci);
    end;
  until (c = 0) or (i >= Length(s));
  case ci of
    1: s1 := Copy(s,oi,i-oi);
    2: s2 := Copy(s,oi,i-oi);
    3: s3 := Copy(s,oi,i-oi);
  end;
  Inc(i);
end;

function TParser.Str2OPZ(s:String):String;
label 1;
var i,i1,j,p,ci,cn: Integer;
    stack: String;
    res,s1,s2,s3,s4,sr: String;
    vr: Boolean;
    c: Char;
    R1,R2,R3:TParserValue;
function Priority(c:Char): Integer;
begin
  case c of
    '(': Priority := 5;
    ')': Priority := 4;
    '=','>','<',ttGe,ttLe,ttNe: Priority := 3;
    '+','-',ttUnMinus,ttUnPlus: Priority := 2;
    '*','/',ttOr,ttAnd,ttNot,ttMod: Priority := 1;
    ttInt,ttFrac,ttRound,ttStr: Priority := 0;
    else Priority := 0;
  end;
end;
procedure ProcessQuotes(var s:String);
var i: Integer;
begin
  if s[1] <> '''' then Exit;
  i := 2;
  if Length(s) > 2 then
    while i <= Length(s) do
    begin
      if (s[i] = '''') and (i < Length(s)) then
      begin
        Insert('''',s,i);
        Inc(i);
      end;
      Inc(i);
    end;
end;
begin
  res := '';
  stack := '';
  i := 1; vr := False;
  while i <= Length(s) do
  begin
    case s[i] of
      '(':
        begin
          stack := '(' + stack;
          vr := False;
        end;
      ')':
        begin
          p := Pos('(', stack);
          res := res + Copy(stack, 1, p - 1);
          stack := Copy(stack, p + 1, Length(stack) - p);
        end;
      '+','-','*','/','>','<','=':
        begin
          if (s[i] = '<') and (s[i + 1] = '>') then
          begin
            Inc(i);
            s[i] := ttNe;
          end else
          if (s[i] = '>') and (s[i + 1] = '=') then
          begin
            Inc(i);
            s[i] := ttGe;
          end else
          if (s[i] = '<') and (s[i + 1] = '=') then
          begin
            Inc(i);
            s[i] := ttLe;
          end;

1:        if not vr then
          begin
            if s[i] = '-' then s[i] := ttUnMinus;
            if s[i] = '+' then s[i] := ttUnPlus;
          end;
          vr := False;
          if stack = '' then stack := s[i] + stack
          else
            if Priority(s[i]) < Priority(stack[1]) then
              stack := s[i] + stack
            else
            begin
              repeat
                res := res + stack[1];
                stack := Copy(stack, 2, Length(stack) - 1);
              until (stack = '') or (Priority(stack[1]) > Priority(s[i]));
              stack := s[i] + stack;
            end;
        end;
      ';': break;
      ' ': ;
      else
      begin
        vr := True;
        s2 := '';
        i1 := i;
        if s[i] = '%' then
        begin
          s2 := '%' + s[i+1];
          Inc(i,2);
        end;
        if s[i] = '''' then
          s2 := s2 + GetString(s,i)
        else if s[i] = '[' then
        begin
          s2 := s2 + '[' + GetBrackedVariable(s,i,j) + ']';
          i := j + 1;
        end
        else
          s2 := s2 + GetIdentify(s, i);
        c := s[i];
        if (Length(s2) > 0) and (s2[1] in ['0'..'9','.',',']) then
          res := res + s2 + ' '
        else
        begin
          s1 := stUpCase(s2);
          if s1 = 'INT' then
          begin
            s[i-1] := ttInt;
            Dec(i);
            goto 1;
          end
          else if s1 = 'FRAC' then
          begin
            s[i-1] := ttFrac;
            Dec(i);
            goto 1;
          end
          else if s1 = 'ROUND' then
          begin
            s[i-1] := ttRound;
            Dec(i);
            goto 1;
          end
          else if s1 = 'OR' then
          begin
            s[i-1] := ttOr;
            Dec(i);
            goto 1;
          end
          else if s1 = 'AND' then
          begin
            s[i-1] := ttAnd;
            Dec(i);
            goto 1;
          end
          else if s1 = 'NOT' then
          begin
            s[i-1] := ttNot;
            Dec(i);
            goto 1;
          end
          else if s1 = 'STR' then
          begin
            s[i-1] := ttStr;
            Dec(i);
            goto 1;
          end
          else if s1 = 'MOD' then
          begin
            s[i-1] := ttMod;
            Dec(i);
            goto 1;
          end
          else if c = '(' then
          begin
            Get3Parameters(s, i, s2, s3, s4);
            res := res + Copy(s, i1, i - i1);
          end
          else res := res + s2 + ' ';
        end;
        Dec(i);
      end;
    end;
    Inc(i);
  end;
  if stack <> '' then res := res + stack;
  Str2OPZ := res;
end;

procedure TParser.Calc(const s:String; var Result:TParserValue);
begin
  CalcOPZ(Str2OPZ(s), Result);
end;

end.
